@@ -36,6 +36,9 @@
 
 #include "egg/egg-dbus.h"
 
+/* Exposed in gkr-operation.h */
+gboolean gkr_inited = FALSE;
+
 static DBusConnection *dbus_connection = NULL;
 G_LOCK_DEFINE_STATIC(dbus_connection);
 
@@ -65,6 +68,30 @@ struct _GkrOperation {
 	GSList *completed;
 };
 
+void
+gkr_operation_init (void)
+{
+	static gsize onced = 0;
+
+	if (g_once_init_enter (&onced)) {
+		gkr_inited = TRUE;
+
+		/*
+		 * Make sure we initialize dbus thread primitives as apps may
+		 * (and do) call g-k in threads. This isn't 100% safe, because
+		 * someone might have been doing dbus calls before without
+		 * initializing threads. However, there is no safe solution to
+		 * this other than just making all such places initialze threads
+		 * (or port to GDBus), so we do it anyway. (GVfs does it too, so
+		 * most apps end up doing it that way.)
+		 */
+
+		dbus_threads_init_default ();
+
+		g_once_init_leave (&onced, 1);
+	}
+}
+
 GkrOperation*
 gkr_operation_ref (GkrOperation *op)
 {
@@ -245,12 +272,20 @@ gkr_operation_complete_later (GkrOperation *op, GnomeKeyringResult res)
 }
 
 static DBusHandlerResult
-on_name_changed_filter (DBusConnection *connection, DBusMessage *message, void *user_data)
+on_connection_filter (DBusConnection *connection,
+                      DBusMessage *message,
+                      void *user_data)
 {
 	const char *object_name;
 	const char *new_owner;
 	const char *old_owner;
 
+	/* org.freedesktop.DBus.Local.Disconnected() */
+	if (dbus_message_is_signal (message, DBUS_INTERFACE_LOCAL, "Disconnected")) {
+		g_warning ("dbus connection disconnected for unknown reason");
+		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+	}
+
 	/* org.freedesktop.DBus.NameOwnerChanged(STRING name, STRING old_owner, STRING new_owner) */
 	if (dbus_message_is_signal (message, DBUS_INTERFACE_DBUS, "NameOwnerChanged") &&
 	    dbus_message_get_args (message, NULL, DBUS_TYPE_STRING, &object_name,
@@ -306,7 +341,7 @@ connect_to_service (void)
 		/* Listen for name owner changed signals */
 		rule = "type='signal',member='NameOwnerChanged',interface='org.freedesktop.DBus'";
 		dbus_bus_add_match (conn, rule, NULL);
-		dbus_connection_add_filter (conn, on_name_changed_filter, NULL, NULL);
+		dbus_connection_add_filter (conn, on_connection_filter, NULL, NULL);
 
 		G_LOCK (dbus_connection);
 		{
@@ -466,7 +501,7 @@ send_with_reply_and_block (DBusConnection *conn, DBusMessage *message,
 		g_return_val_if_reached (NULL);
 
 	if (pending == NULL) {
-		gkr_debug ("couldn't send message");
+		gkr_debug ("couldn't send message, dbus connection disconnected");
 		return NULL;
 	}
 
